#define UNICODE
#define _UNICODE
#include <windows.h>
#include <crtdbg.h>
#include <string>
#include <vector>
#include "PEFile.h"
#include "LoadFile.h"


//=============================================================================


//! RXgN^BɊJǂ GetLoadBase() ŊmFB
PEFile::PEFile(const wchar_t* i_file_name)
: m_need_to_free(true),
  m_load_base(NULL),
  m_dos_header_p(NULL),
  m_nt_headers_p(NULL),
  m_section_table_p(NULL)
{
  do
  {
    // C[Wt@CɃ}bvB
    m_load_base = MapImage(i_file_name);
    if( m_load_base == NULL )
    {
      _RPTF1(_CRT_WARN, "[%S]̃[hɎs܂B", i_file_name);
      break;
    }
    _RPTF1(_CRT_WARN, "m_load_base: %p\n", m_load_base);

    // DOS wb_ւ̃|C^擾B
    m_dos_header_p = reinterpret_cast<IMAGE_DOS_HEADER*>(m_load_base);
    
    // NT wb_ւ̃|C^擾B
    m_nt_headers_p = reinterpret_cast<IMAGE_NT_HEADERS32*>(
                         m_load_base + m_dos_header_p->e_lfanew
                     );
      
    // ZNVe[uւ̃|C^擾B
    m_section_table_p = reinterpret_cast<IMAGE_SECTION_HEADER*>(
                          m_nt_headers_p + 1
                        );
  }
  while( 0 );
}


//=============================================================================


//! fXgN^
PEFile::~PEFile()
{
  if( m_need_to_free && m_load_base )
  {
    VirtualFree(m_load_base, 0, MEM_RELEASE);
    m_load_base = NULL;
  }
}


//=============================================================================


//! [hAhX擾B
BYTE* PEFile::GetLoadBase()
{
  return m_load_base;
}


//=============================================================================


//! DOS wb_ւ̃|C^擾B
IMAGE_DOS_HEADER* PEFile::GetDosHeaderP()
{
  return GetLoadBase() ? m_dos_header_p : NULL;
}


//=============================================================================


//! NT wb_ւ̃|C^擾B
IMAGE_NT_HEADERS32* PEFile::GetNtHeadersP()
{
  return GetLoadBase() ? m_nt_headers_p : NULL;
}


//=============================================================================


//! t@Cwb_ւ̃|C^擾B
IMAGE_FILE_HEADER* PEFile::GetFileHeaderP()
{
  return GetLoadBase() ? &m_nt_headers_p->FileHeader : NULL;
}


//=============================================================================


//! IvViwb_ւ̃|C^擾B
IMAGE_OPTIONAL_HEADER32* PEFile::GetOptionalHeaderP()
{
  return GetLoadBase() ? &m_nt_headers_p->OptionalHeader : NULL;
}


//=============================================================================


//! i_index Ԗڂ̃ZNVwb_ւ̃|C^擾B
IMAGE_SECTION_HEADER* PEFile::GetSectionHeader(int i_index)
{
  // i_index ZNVɎ܂Ă邩mFB
  if( GetLoadBase() && 0 <= i_index && i_index < GetFileHeaderP()->NumberOfSections )
    return &m_section_table_p[i_index];
  else
    return NULL;
}


//=============================================================================


//! i_index Ԗڂ̃f[^fBNgGgwf[^ւ̃|C^擾B
BYTE* PEFile::GetDirEntryDataP(int i_index)
{
  if( GetLoadBase()
  &&  0 <= i_index && i_index < IMAGE_NUMBEROF_DIRECTORY_ENTRIES
  &&  GetOptionalHeaderP()->DataDirectory[i_index].VirtualAddress
  &&  GetOptionalHeaderP()->DataDirectory[i_index].Size )
  {
    return m_load_base + GetOptionalHeaderP()->DataDirectory[i_index].VirtualAddress;
  }
  else
  {
    return NULL;
  }
}


//=============================================================================


//! i_index Ԗڂ̃f[^fBNgGgwf[^ւ̃TCY擾B
DWORD PEFile::GetDirEntryDataSize(int i_index)
{
  if( GetLoadBase() && 0 <= i_index && i_index < IMAGE_NUMBEROF_DIRECTORY_ENTRIES )
    return GetOptionalHeaderP()->DataDirectory[i_index].Size;
  else
    return 0;
}


//=============================================================================


//! C|[gfBNgւ̃|C^擾B
IMAGE_IMPORT_DESCRIPTOR* PEFile::GetImportDirP()
{
  return reinterpret_cast<IMAGE_IMPORT_DESCRIPTOR*>(
           GetDirEntryDataP(IMAGE_DIRECTORY_ENTRY_IMPORT)
         );
}


//=============================================================================


//! C|[gĂ DLL 擾B
std::vector<const char*> PEFile::GetImportDllNames()
{
  std::vector<const char*>  retval;
  IMAGE_IMPORT_DESCRIPTOR*  imp_descs_p = GetImportDirP();
  
  if( imp_descs_p )
  {
    for(int i = 0; imp_descs_p[i].FirstThunk != 0; i++)
    {
      retval.push_back(GetDataP<const char*>(imp_descs_p[i].Name));
    }
  }
  
  return retval;
}


//=============================================================================


//! w肵 DLL C|[gĂV{擾B
std::vector<ImportSymbol> PEFile::GetImportSymbols(const char* i_dll_name)
{
  std::vector<ImportSymbol>   retval;
  IMAGE_IMPORT_DESCRIPTOR*    imp_descs_p = GetImportDirP();

  // C|[gl[e[uĂ邩APEFile NXOŃ[h
  // C[Wt@Ĉݗ񋓉\B
  if( imp_descs_p && (imp_descs_p[0].OriginalFirstThunk || m_need_to_free) )
  {
    for(int i = 0; imp_descs_p[i].FirstThunk != 0; i++)
    {
      const char* dll_name = GetDataP<const char*>(imp_descs_p[i].Name);
      if( _stricmp(dll_name, i_dll_name) == 0 )  // w肳ꂽ DLL 𔭌!
      {
        // C|[gAhXe[uւ̃|C^擾B
        DWORD* iat_p = GetDataP<DWORD*>(imp_descs_p[i].FirstThunk);
        // C|[gl[e[uւ̃|C^擾B
        // INT ݂Ȃꍇ͑ IAT ւ̃|C^gpB
        DWORD* int_p = imp_descs_p[i].OriginalFirstThunk
                     ? GetDataP<DWORD*>(imp_descs_p[i].OriginalFirstThunk)
                     :  iat_p;

        // C|[gV{
        for(int table_index = 0; int_p[table_index] != 0; table_index++)
        {
          // INT ̃R[hC[WɎ܂Ă邩mFB
          if( int_p[table_index] < GetOptionalHeaderP()->SizeOfImage )
          {
            ImportSymbol  symbol = { 0 };

            symbol.iat_entry_rva = imp_descs_p[i].FirstThunk
                                 + sizeof(DWORD) * table_index;
            symbol.iat_entry_p   = reinterpret_cast<void**>(&iat_p[table_index]);

            // ɂC|[g
            if( IMAGE_SNAP_BY_ORDINAL32(int_p[table_index]) ) 
            {
              symbol.ordinal = static_cast<WORD>(IMAGE_ORDINAL32(int_p[table_index]));
            }
            // OɂC|[g
            else
            {
              IMAGE_IMPORT_BY_NAME*
                by_name_p = GetDataP<IMAGE_IMPORT_BY_NAME*>(int_p[table_index]);
              symbol.hint = by_name_p->Hint;
              symbol.name = reinterpret_cast<char*>(by_name_p->Name);
            }

            retval.push_back(symbol);
          }
        }
      }
    }
  }
  
  return retval;
}


//=============================================================================


//! C[Wt@CɃ}bsOB
BYTE* PEFile::MapImage(const wchar_t* i_file_name)
{
  BYTE*     retval          = NULL;
  
  do
  {
    // t@CŜǂݍށB
    std::vector<BYTE> file_data;
    if( LoadFile(i_file_name, file_data) == false )
      break;

    // DOS wb_̎擾ƊmFB
    IMAGE_DOS_HEADER* dos_header_p = reinterpret_cast<IMAGE_DOS_HEADER*>(
                                       &file_data[0]
                                     );
    if( dos_header_p->e_magic != *reinterpret_cast<WORD*>("MZ")
    ||  dos_header_p->e_lfanew == 0 )
    {
      _RPTF0(_CRT_WARN, "s\t@Cł͂܂B\n");
      break;
    }
    
    // NT wb_̎擾ƊmFB
    IMAGE_NT_HEADERS32* nt_headers_p = reinterpret_cast<IMAGE_NT_HEADERS*>(
                                         &file_data[dos_header_p->e_lfanew]
                                       );
    if( nt_headers_p->Signature != *reinterpret_cast<DWORD*>("PE\0\0")
    ||  nt_headers_p->OptionalHeader.Magic != IMAGE_NT_OPTIONAL_HDR32_MAGIC )
    {
      _RPTF0(_CRT_WARN, "PE t@Cł͂܂B\n");
      break;
    }
    
    // ZNVe[uւ̃|C^擾B
    IMAGE_SECTION_HEADER* section_table_p = reinterpret_cast<IMAGE_SECTION_HEADER*>(
                                              nt_headers_p + 1
                                            );
    
    // SizeOfImage mہB
    retval  = reinterpret_cast<BYTE*>(
                VirtualAlloc(
                  NULL,
                  nt_headers_p->OptionalHeader.SizeOfImage,
                  MEM_COMMIT,
                  PAGE_READWRITE
                )
              );
    if( retval == NULL )
    {
      _RPTF1(_CRT_WARN, "̊蓖[%dbytes]Ɏs܂B\n",
             nt_headers_p->OptionalHeader.SizeOfImage
             );
      break;
    }

    // wb_Rs[
    memcpy(retval, &file_data[0], nt_headers_p->OptionalHeader.SizeOfHeaders);
    // ZNVf[^Rs[
    for(DWORD i = 0; i < nt_headers_p->FileHeader.NumberOfSections; i++)
    {
      // ZNVf[^ꍇB
      if( section_table_p[i].PointerToRawData )
      {
        memcpy(&retval[section_table_p[i].VirtualAddress],
               &file_data[section_table_p[i].PointerToRawData],
               section_table_p[i].SizeOfRawData
              );
      }
    }
  }
  while( 0 );
  
  return retval;
}


//=============================================================================


